/* -*-mode: c; indent-tabs-mode: nil; c-basic-offset: 2; -*-
 */

/**
 * mod_setenvifplus.c: Apache httpd module to set environment variables
 * See http://modsetenvifplus.sourceforge.net/ for further
 * details about mod_setenvifplus.
 *
 * Copyright (C) 2020 Pascal Buchbinder
 *
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/************************************************************************
 * Version
 ***********************************************************************/
static const char revision[] = "$Id: mod_setenvifplus.c 177 2020-07-01 05:12:41Z pbuchbinder $";
static const char g_revision[] = "0.40";

/************************************************************************
 * Includes
 ***********************************************************************/

#include <arpa/inet.h>

/* openssl */
#include <openssl/crypto.h>
#include <openssl/dh.h>
#include <openssl/bn.h>
#include <openssl/rand.h>
#include <openssl/evp.h>
#include <openssl/err.h>

/* apr */
#include "apr.h"
#include "apr_strings.h"
#include "apr_strmatch.h"
#include "apr_lib.h"
#include "apr_base64.h"

#define APR_WANT_STRFUNC
#include "apr_want.h"

/* apache */
//#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"
#include <util_filter.h>
#include "util_md5.h"

/************************************************************************
 * defines
 ***********************************************************************/
#define SP_LOG_PFX(id)  "mod_setenvifplus("#id"): "
#define SP_COOKIE_KEY "SP_COOKIE_KEY"
#define SP_KEY "SP_KEY"
#define SP_RAND_SIZE 18

enum sp_special {
  SP_SPECIAL_NOT,
  SP_SPECIAL_REMOTE_ADDR,
  SP_SPECIAL_REMOTE_NET,
  SP_SPECIAL_REMOTE_HOST,
  SP_SPECIAL_REQUEST_URI,
  SP_SPECIAL_REQUEST_QUERY,
  SP_SPECIAL_REQUEST_USER,
  SP_SPECIAL_REQUEST_METHOD,
  SP_SPECIAL_REQUEST_PROTOCOL,
  SP_SPECIAL_SERVER_ADDR,
  SP_SPECIAL_SERVER_PORT,
  SP_SPECIAL_RESPONSE_STATUS
};

enum sp_action {
  SP_ACTION_SET,
  SP_ACTION_UNSET,
  SP_ACTION_CHANGE,
  SP_ACTION_ADD
};

enum sp_cmp {
  SP_CMP_EQ,
  SP_CMP_NE,
  SP_CMP_GT,
  SP_CMP_LT
};

// Apache 2.4 compat
#if (AP_SERVER_MINORVERSION_NUMBER == 4)
#define SF_CONN_REMOTEIP(r) r->connection->client_ip
#define SF_CONN_REMOTE_ADDR(r) r->connection->client_addr
#else
#define SF_CONN_REMOTEIP(r) r->connection->remote_ip
#define SF_CONN_REMOTE_ADDR(r) r->connection->remote_addr
#endif

/************************************************************************
 * structures
 ***********************************************************************/
typedef struct {
  const char *name;             /* attribute name */
  ap_regex_t *pnamereg;         /* compiled attribute name (optional regex) */
  const char *regex;            /* regex to match against */
  ap_regex_t *preg;             /* compiled regex */
  const apr_strmatch_pattern *pattern; /* non-regex pattern to match */
  const char *variable;         /* env var name to set (or unset) */
  const char *value;            /* env var value to set */ 
  int valueSet;                 /* indicates, that a value has been configured by the user*/
  enum sp_special special_type; /* is it a "special" attribute ? */
  int icase;                    /* ignoring case? */
} sp_std_entry_t;

typedef struct {
  const char *name;
  ap_regex_t *preg;             /* compiled regex (for SP_ACTION_CHANGE only) */
  const char *value;
  const char *condition;
  enum sp_action action;
} sp_hdr_entry_t;

typedef struct {
  enum sp_cmp cmp;
  const char *left;
  const char *right;
  const char *variable;
  const char *value;
} sp_cmp_entry_t;

typedef struct {
  const char *header;
  ap_regex_t *preg;
  const char *value;
} sp_usr_entry_t;

typedef struct {
  int code;
  const char *var;
} sp_status_entry_t;

typedef struct {
  const char *header;
  const char *var;
} sp_hash_entry_t;

typedef struct {
  const char *filtername;
  const char *var;
  int notset;
} sp_filter_entry_t;

typedef struct {
  char *path;
  apr_array_header_t *std_conditionals;
  apr_array_header_t *std_late_conditionals;
  apr_array_header_t *std_req_query;
  apr_array_header_t *std_req_changequery;
  apr_array_header_t *std_res_conditionals;
  apr_array_header_t *std_req_header;
  apr_array_header_t *std_late_req_header;
  apr_array_header_t *std_res_header;
  apr_array_header_t *user;
  apr_array_header_t *std_removepattern_req_header;
  apr_array_header_t *status_var;
  apr_array_header_t *outfilter;
  apr_array_header_t *cookie_names;
  apr_array_header_t *header_hash;
  apr_array_header_t* cmp;
  apr_array_header_t* cmp_late;
} sp_config_t;

/* functions to replace patterns */
typedef struct {
  char *identifier;                     /** name of the function inc. opening
                                            bracket, e.g. "e64(" */
  int len;
  char close;                           /** end marker, e.g. ')' */
  char *(*func)(request_rec *, char *); /** pointer to the function */
} sp_function_t;

typedef struct sp_url_escape_seq {
  char c;
  char *esc;
  apr_size_t len;
} sp_url_escape_seq_t;

/************************************************************************
 * globals
 ***********************************************************************/
module AP_MODULE_DECLARE_DATA setenvifplus_module;
#define SP_ICASE_MAGIC  ((void *)(&setenvifplus_module))

#ifdef SP_VAR_PFX_24
#define SP_VAR_PFX "%{"
#else
#define SP_VAR_PFX "${"
#endif
static char *m_sp_var_pfx = SP_VAR_PFX;

/************************************************************************
 * private
 ***********************************************************************/

/**
 * Returns the client's network (IPv4/24 or IPv6/64) number.
 */
static char *sp_get_remote_net(request_rec *r) {
  apr_sockaddr_t *remote_addr = SF_CONN_REMOTE_ADDR(r);
  if(remote_addr->family == AF_INET6) {
    char *dst = apr_pcalloc(r->pool, INET6_ADDRSTRLEN);
    char *ret;
    apr_uint32_t *addr = (apr_uint32_t *)remote_addr->ipaddr_ptr;
    apr_uint32_t dest[4];
    if(IN6_IS_ADDR_V4MAPPED((struct in6_addr *)remote_addr->ipaddr_ptr)) {
      dest[0] = addr[0];
      dest[1] = addr[1];
      dest[2] = addr[2];
      dest[3] = addr[3] & 0x00ffffff;
    } else {
      dest[0] = addr[0];
      dest[1] = addr[1];
      dest[2] = 0;
      dest[3] = 0;
    }
    ret = (char *)inet_ntop(AF_INET6, dest, dst, INET6_ADDRSTRLEN);
    return ret;
  } else {
    char *dst = apr_pcalloc(r->pool, INET_ADDRSTRLEN);
    char *ret;
    apr_uint32_t *addr = (apr_uint32_t *)remote_addr->ipaddr_ptr;
    apr_uint32_t dest[1];
    dest[0] = addr[0] & 0x00ffffff;
    ret = (char *)inet_ntop(AF_INET, dest, dst, INET_ADDRSTRLEN);
    return ret;
  }
}

/**
 * Decryptes encrypted and base64 encoded string. See sp_enc64().
 *
 * @param r
 * @param str String to decrypt
 * @return Decrypted string or NULL on error
 */
static char *sp_dec64(request_rec *r, const char *str, unsigned char *key) {
#if OPENSSL_VERSION_NUMBER < 0x10100000L
  EVP_CIPHER_CTX cipher_ctx;
  EVP_CIPHER_CTX *cipher_ctx_p = &cipher_ctx;
#else
  EVP_CIPHER_CTX *cipher_ctx_p;
#endif
  unsigned char *hashIn;
  unsigned char hash[APR_MD5_DIGESTSIZE];
  apr_md5_ctx_t md5;
  int len = 0;
  int buf_len = 0;
  unsigned char *buf;
  char *dec = (char *)apr_palloc(r->pool, 1 + apr_base64_decode_len(str));
  int dec_len = apr_base64_decode(dec, str);
  buf = apr_pcalloc(r->pool, dec_len);

#if OPENSSL_VERSION_NUMBER < 0x10100000L
  EVP_CIPHER_CTX_init(cipher_ctx_p);
#else
  cipher_ctx_p = EVP_CIPHER_CTX_new();
#endif
  EVP_DecryptInit(cipher_ctx_p, EVP_des_ede3_cbc(), key, NULL);
  if(!EVP_DecryptUpdate(cipher_ctx_p, (unsigned char *)&buf[buf_len], &len,
                        (const unsigned char *)dec, dec_len)) {
    goto failed;
  }
  buf_len+=len;
  if(!EVP_DecryptFinal(cipher_ctx_p, (unsigned char *)&buf[buf_len], &len)) {
    goto failed;
  }
  buf_len+=len;
  
  if(buf_len < (SP_RAND_SIZE + APR_MD5_DIGESTSIZE)) {
    goto failed;
  }
  if(buf[SP_RAND_SIZE-1] != 'A' || buf[SP_RAND_SIZE-2] != 'z') {
    goto failed;
  }

  hashIn = &buf[SP_RAND_SIZE];
  buf = &buf[SP_RAND_SIZE + APR_MD5_DIGESTSIZE];
  buf_len = buf_len - SP_RAND_SIZE - APR_MD5_DIGESTSIZE;

  apr_md5_init(&md5);
  apr_md5_update(&md5, buf, buf_len);
  apr_md5_final(hash, &md5);

  if(memcmp(hash, hashIn, APR_MD5_DIGESTSIZE) != 0) {
    goto failed;
  }

#if OPENSSL_VERSION_NUMBER < 0x10100000L
  EVP_CIPHER_CTX_cleanup(cipher_ctx_p);
#else
  EVP_CIPHER_CTX_free(cipher_ctx_p);
#endif

  return apr_pstrndup(r->pool, (char *)buf, buf_len);
     
 failed:
#if OPENSSL_VERSION_NUMBER < 0x10100000L
  EVP_CIPHER_CTX_cleanup(cipher_ctx_p);
#else
  EVP_CIPHER_CTX_free(cipher_ctx_p);
#endif
  ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, 0, r,
                SP_LOG_PFX(010)"Failed to decrypt data. (%s)", str ? str : "");
  return NULL;
}

/**
 * Encrypts and base64 encodes a string. See sp_dec64().
 *
 * @param r
 * @param str String to encrypt
 * @return Encrypted string.
 */
static char *sp_enc64(request_rec *r, const char *str, unsigned char *key) {
  char *e;
#if OPENSSL_VERSION_NUMBER < 0x10100000L
  EVP_CIPHER_CTX cipher_ctx;
  EVP_CIPHER_CTX *cipher_ctx_p = &cipher_ctx;
#else
  EVP_CIPHER_CTX *cipher_ctx_p;
#endif
  unsigned char hash[APR_MD5_DIGESTSIZE];
  apr_md5_ctx_t md5;
  int len = 0;
  int buf_len = 0;
  int str_len = strlen(str);
  int max_buf_len = SP_RAND_SIZE + 
    str_len + 
    EVP_CIPHER_block_size(EVP_des_ede3_cbc()) +
    APR_MD5_DIGESTSIZE;
  unsigned char *buf = apr_pcalloc(r->pool, max_buf_len);
  unsigned char *rand = apr_pcalloc(r->pool, SP_RAND_SIZE);
#if APR_HAS_RANDOM
  if(apr_generate_random_bytes(rand, SP_RAND_SIZE) != APR_SUCCESS) {
    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                  SP_LOG_PFX(030)"Can't generate random data.");
  }
#else
  if(!RAND_bytes(rand, SP_RAND_SIZE)) {
    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                  SP_LOG_PFX(030)"Can't generate random data.");
  }
#endif
  apr_md5_init(&md5);
  apr_md5_update(&md5, str, str_len);
  apr_md5_final(hash, &md5);
  rand[SP_RAND_SIZE-1] = 'A';
  rand[SP_RAND_SIZE-2] = 'z';
#if OPENSSL_VERSION_NUMBER < 0x10100000L
  EVP_CIPHER_CTX_init(cipher_ctx_p);
#else
  cipher_ctx_p = EVP_CIPHER_CTX_new();
#endif
  EVP_EncryptInit(cipher_ctx_p, EVP_des_ede3_cbc(), key, NULL);
  if(!EVP_EncryptUpdate(cipher_ctx_p, &buf[buf_len], &len,
                        rand, SP_RAND_SIZE)) {
    goto failed;
  }
  buf_len+=len;
  if(!EVP_EncryptUpdate(cipher_ctx_p, &buf[buf_len], &len,
                        hash, APR_MD5_DIGESTSIZE)) {
    goto failed;
  }
  buf_len+=len;
  if(!EVP_EncryptUpdate(cipher_ctx_p, &buf[buf_len], &len,
                        (const unsigned char *)str, str_len)) {
    goto failed;
  }
  buf_len+=len;
  if(!EVP_EncryptFinal(cipher_ctx_p, &buf[buf_len], &len)) {
    goto failed;
  }
  buf_len+=len;
#if OPENSSL_VERSION_NUMBER < 0x10100000L
  EVP_CIPHER_CTX_cleanup(cipher_ctx_p);
#else
  EVP_CIPHER_CTX_free(cipher_ctx_p);
#endif

  e = (char *)apr_pcalloc(r->pool, 1 + apr_base64_encode_len(buf_len));
  len = apr_base64_encode(e, (const char *)buf, buf_len);
  e[len] = '\0';
  return e;

 failed:
#if OPENSSL_VERSION_NUMBER < 0x10100000L
  EVP_CIPHER_CTX_cleanup(cipher_ctx_p);
#else
  EVP_CIPHER_CTX_free(cipher_ctx_p);
#endif
  ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, 0, r,
                SP_LOG_PFX(011)"Failed to encrypt data.");
  return "";
}

/**
 * Similar to strstr but restricting the length of s1 (if not null terminated).
 *
 * @param s1 String to search in
 * @param s2 Pattern to ind
 * @param len Length of s1
 * @return pointer to the beginning of the substring s2 within s1, or NULL
 *         if the substring is not found
 */
static char *sp_strnstr(const char *s1, const char *s2, int len) {
  const char *e1 = &s1[len-1];
  const char *p, *q;
  for (; *s1 && (s1 <= e1); s1++) {
    p = s1, q = s2;
    while(*q && *p && (q <= e1)) {
      if (*q != *p) {
        break;
      }
      p++, q++;
    }
    if(q > e1) {
      return NULL;
    }
    if(*q == 0) {
      return (char *)s1;
    }
  }
  return 0;
}

static int sp_isnum(const char *x) {
  const char *p = x;
  if(x == NULL || x[0] == 0) {
    return 0;
  }
  while(p && p[0]) {
    if(!apr_isdigit(p[0])) {
      return 0;
    }
    p++;
  }
  return 1;
}
/**
 * Retuns the hex value of the character
 *
 * @param x
 * @return hex value
 */
static int sp_hex2c(const char *x) {
  int i, ch;
  ch = x[0];
  if (apr_isdigit(ch)) {
    i = ch - '0';
  }else if (apr_isupper(ch)) {
    i = ch - ('A' - 10);
  } else {
    i = ch - ('a' - 10);
  }
  i <<= 4;
  
  ch = x[1];
  if (apr_isdigit(ch)) {
    i += ch - '0';
  } else if (apr_isupper(ch)) {
    i += ch - ('A' - 10);
  } else {
    i += ch - ('a' - 10);
  }
  return i;
}

#define SP_ISHEX(x) (((x >= '0') && (x <= '9')) ||  \
                     ((x >= 'a') && (x <= 'f')) ||  \
                     ((x >= 'A') && (x <= 'F')))


/**
 * URL escaping/encoding (%xx, '+')
 *
 * @param pool To allocate encoded string from.
 * @param x String to encode
 * @retrun encoded string
 */
static char *sp_url_encoding(apr_pool_t *pool, const char *x) {
  int i = 0;
  int k = 0;
  int len = strlen(x);
  char *c = apr_pcalloc(pool, (3 * len) + 1);
  unsigned char *in = (unsigned char *)x;
  while(in[i]) {
    if(strchr("?\n /;%=\"'.:@\\&+", in[i]) != NULL) {
      sprintf(&c[k], "%%%02x", in[i]);
      k+=3;
    } else {
      c[k] = in[i];
      k++;
    }
    i++;
  }
  c[k] = '\0';
  return c;
}

/**
 * URL unescaping/decoding (%xx, '+')
 *
 * @param x String to decode
 * @param error Error message if decoding fails
 * @return Length of the decoded string
 */
static int sp_url_decoding(char *x, int *error) {
  int i, j, ch;
  if(x == 0) {
    return 0;
  }
  if(x[0] == '\0') {
    return 0;
  }
  for(i = 0, j = 0; x[i] != '\0'; i++, j++) {
    ch = x[i];
    if(ch == '%') {
      if(SP_ISHEX(x[i + 1]) && SP_ISHEX(x[i + 2])) {
        ch = sp_hex2c(&x[i + 1]);
        i += 2;
      } else {
        (*error)++;
      }
    } else if(ch == '+') {
      ch = ' ';
    }
    x[j] = ch;
  }
  x[j] = '\0';
  return j;
}

/**
 * If a Header name contains characters other than:
 *    -,_,[A-Z\, [a-z] and [0-9].
 * assume the header name is a regular expression.
 *
 * @param p
 * @param name
 * @return 1 if it is a regular expression or 0 if it can be treaten as literal string 
 */
static int sp_is_header_regex(apr_pool_t *p, const char* name) {
  ap_regex_t *preg = ap_pregcomp(p, "^[-A-Za-z0-9_]*$", (AP_REG_EXTENDED|AP_REG_NOSUB));
  ap_assert(preg != NULL);
  if(ap_regexec(preg, name, 0, NULL, 0)) {
    return 1;
  }
  return 0;
}

/**
 * If the input string does not take advantage of regular
 * expression metacharacters, return a pointer to an equivalent
 * string that can be searched using apr_strmatch().  (The
 * returned string will often be the input string.  But if
 * the input string contains escaped characters, the returned
 * string will be a copy with the escapes removed.)
 */
static const char *sp_non_regex_pattern(apr_pool_t *p, const char *s) {
  const char *src = s;
  int escapes_found = 0;
  int in_escape = 0;
  
  while (*src) {
    switch (*src) {
    case '^':
    case '.':
    case '$':
    case '|':
    case '(':
    case ')':
    case '[':
    case ']':
    case '*':
    case '+':
    case '?':
    case '{':
    case '}':
      if(!in_escape) {
        return NULL;
      }
      in_escape = 0;
      break;
    case '\\':
      if(!in_escape) {
        in_escape = 1;
        escapes_found = 1;
      } else {
        in_escape = 0;
      }
      break;
    default:
      if(in_escape) {
        return NULL;
      }
      break;
    }
    src++;
  }
  if(!escapes_found) {
    return s;
  } else {
    char *unescaped = (char *)apr_palloc(p, src - s + 1);
    char *dst = unescaped;
    src = s;
    do {
      if(*src == '\\') {
        src++;
      }
    } while((*dst++ = *src++));
    return unescaped;
  }
}

/**
 * Replaces <start>var<end> (e.g. ${var}) entries by the value of the
 * variable within the vars table.
 *
 * @param r
 * @param start_marker Defines thow the vairable beginns, e,g "${" or "b64("
 * @param start_marker_len Lenth of the marker string
 * @param end_marker Counterpart to the start marker, e.g. "}" or ")"
 * @param string String to respove variables
 * @param func Function to replace the placeholder
 * @param envvar Name of the variable to replace the value for (itself, optional)
 * @returns 1 (success) if all start markers could be replaced
 */
static int sp_reslove_variable(request_rec *r,
                               const char *start_marker,
                               int start_marker_len,
                               const char end_marker,
                               char **string,
                               char *(*func)(request_rec *, char *),
                               const char *envvar) {
  char startmarker = '\0';
  char endmarker = '\0';
  int i;
  int start;
  int line_end;
  char *var_name;
  char *new_line = *string;
  char *line = *string;
  const char *val;
  
  /* functions may enclose strings which contains brackets
       var=Mozilla/5.0 (Windows NT 5.1; rv:2.0.1) Gecko/20110000 Firefox/3.6
                       ^                        ^
       header=base64(${var})
     => if we found another start marker and that we should skip the next end marker */
  if(sp_strnstr(start_marker, "(", start_marker_len)) {
    startmarker = '(';
    endmarker = ')';
  }
 once_again:
  i = 0;
  while(line[i] != 0) {
    int found_nested = 0;
    if(strncmp(&line[i], start_marker, start_marker_len) == 0) {
      line_end = i;
      i+=start_marker_len;
      start = i;
      // search the end_marker
      while((line[i] != 0) &&
            !((line[i] == end_marker) && (found_nested == 0))) {
        if(startmarker) {
          if(line[i] == startmarker) {
            found_nested++;
          }
          if((line[i] == endmarker) && found_nested) {
            found_nested--;
          }
        }
        i++;
      }
      // did we miss the end? because of more start than end markers?
      if(startmarker) {
        if(line[i] == '\0' && line[i-1] == end_marker) {
          i--;
        }
      }
      if(line[i] != end_marker) {
        /* no end found */
        break;
      } else {
        var_name = apr_pstrndup(r->pool, &line[start], i - start);
        val = func(r, var_name); 
        if(!val && envvar && (strcmp(var_name, envvar) == 0)) {
          /* resolving the variable itself may fail (use might want to 
             concatenate variables)
             => replace it by an empty value */
          val = apr_pstrdup(r->pool, "");
        }
        if(val) {
          line[line_end] = 0;
          i++;
          new_line = apr_pstrcat(r->pool, line, val, &line[i], NULL);
          line = new_line;
          goto once_again;
        }      
      }
    }
    i++;
  }
  if(!new_line[0] || strstr(new_line, start_marker)) {
    return 0;
  }
  *string = new_line;
  return 1;
}

/**
 * Function: replaces the env variable by its value.
 *
 * @param r
 * @param variable
 * @return
 */
static char *sp_func_variable(request_rec *r, char *variable) {
  return apr_pstrdup(r->pool, apr_table_get(r->subprocess_env, variable));
  
}

/**
 * Function: URL encoding
 *
 * @param r
 * @param string String to encode
 * @return Encoded string
 */
static char *sp_func_eURL(request_rec *r, char *string) {
  return sp_url_encoding(r->pool, string);
}

/**
 * Function: URL decoding
 *
 * @param r
 * @param string String to decode
 * @return Decoded string
 */
static char *sp_func_dURL(request_rec *r, char *string) {
  int error;
  char *c = apr_pstrdup(r->pool, string);
  sp_url_decoding(c, &error);
  return c;
}

/**
 * Function: encryption
 *
 * @param r
 * @param string String to encode
 * @return Encoded string
 */
static char *sp_func_enc(request_rec *r, char *string) {
  const char *keyStr = apr_table_get(r->subprocess_env, SP_KEY);
  unsigned char key[EVP_MAX_KEY_LENGTH];
  if(keyStr == NULL) {
    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                  SP_LOG_PFX(031)"$enc() can't encrypt string, no key available."
                  " "SP_KEY" variable is missing!");
    return "";
  }
  EVP_BytesToKey(EVP_des_ede3_cbc(), EVP_sha1(), NULL, (unsigned char *)keyStr,
                 strlen(keyStr), 1, key, NULL);
  
  return sp_enc64(r, string, key);
}

/**
 * Function: decryption
 *
 * @param r
 * @param string String to decrypt
 * @return Decrypted string
 */
static char *sp_func_dec(request_rec *r, char *string) {
  const char *keyStr = apr_table_get(r->subprocess_env, SP_KEY);
  unsigned char key[EVP_MAX_KEY_LENGTH];
  if(keyStr == NULL) {
    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                  SP_LOG_PFX(031)"$enc() can't encrypt string, no key available."
                  " "SP_KEY" variable is missing!");
    return "";
  }
  EVP_BytesToKey(EVP_des_ede3_cbc(), EVP_sha1(), NULL, (unsigned char *)keyStr,
                 strlen(keyStr), 1, key, NULL);
  return sp_dec64(r, string, key);
}


/**
 * Function: base64 encoding
 *
 * @param r
 * @param string String to encode
 * @return Encoded string
 */
static char *sp_func_e64(request_rec *r, char *string) {
  return ap_pbase64encode(r->pool, string);
}

/**
 * Function: base64 decoding
 *
 * @param r
 * @param string String to decode
 * @return Decoded string
 */
static char *sp_func_d64(request_rec *r, char *string) {
  return ap_pbase64decode(r->pool, string);
}

/**
 * Function: creates radom characters
 *
 * @param r
 * @param num Number of bytes to return
 * @return Random string
 */
static char *sp_func_RND(request_rec *r, char *num) {
  char *buf;
  apr_int64_t len = apr_atoi64(num);
  unsigned char *secret = apr_pcalloc(r->pool, len);
#if APR_HAS_RANDOM
  if(apr_generate_random_bytes(secret, len) != APR_SUCCESS) {
    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                  SP_LOG_PFX(030)"$RND() can't generate random data.");
  }
#else
  if(!RAND_bytes(secret, len)) {
    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                  SP_LOG_PFX(030)"$RND() can't generate random data.");
  }
#endif
  buf = (char *)apr_pcalloc(r->pool, 1 + apr_base64_encode_len(len));
  apr_base64_encode(buf, (const char *)secret, len);
  buf[len] = '\0';
  return buf;
}

/**
 * List of registered functions
 */
static sp_function_t sp_function_list[]= {
  { "$e64(",  5, ')', &sp_func_e64 },
  { "$d64(",  5, ')', &sp_func_d64 },
  { "$dURL(", 6, ')', &sp_func_dURL },
  { "$eURL(", 6, ')', &sp_func_eURL },
  { "$enc(",  5, ')', &sp_func_enc },
  { "$dec(",  5, ')', &sp_func_dec },
  { "$RND(",  5, ')', &sp_func_RND },
  { NULL,     0, 0,   NULL }
};

/**
 * Implements the kown functions, e.g. base64 encoding
 *
 * @param r
 * @param replaced
 * @return Resolved string
 */
static char *sp_process_functions(request_rec *r, const char *replaced) {
  sp_function_t *e = sp_function_list;
  char *string = apr_pstrdup(r->pool, replaced);
  while(e && e->identifier) {
    sp_reslove_variable(r, e->identifier, e->len, e->close, &string, e->func, NULL);
    e++;
  }
  return string;
}

/**
 * Resolves the variable value and (un-)sets it.
 *
 * @param t
 * @param b Rule entry
 * @param val String value to process / resolve. This value is added
 *        as the environment variable as configured by the rule entry b.
 */
static void sp_match_var(request_rec *r, sp_std_entry_t *b, const char *val) {
  if(val) {
    ap_regmatch_t regm[AP_MAX_REG_MATCH];
    apr_size_t val_len = strlen(val);
    if((b->pattern && apr_strmatch(b->pattern, val, val_len)) ||
       (!b->pattern && !ap_regexec(b->preg, val, AP_MAX_REG_MATCH, regm, 0))) {
      if(b->variable[0] == '!') {
        apr_table_unset(r->subprocess_env, &b->variable[1]);
      } else {
        char *replaced = NULL;
        if(!b->pattern) {
          /* 1) respolve sub regex */
          replaced = ap_pregsub(r->pool, b->value, val, AP_MAX_REG_MATCH, regm);
        } else {
          /* literal string (no $1 .. $9) */
          replaced = apr_pstrdup(r->pool, b->value);
        }
        if(replaced) {
          /* 2) sp_reslove_variable() resolves all ${xxx} variables */
          if(sp_reslove_variable(r, m_sp_var_pfx, 2, '}', &replaced,
                                 &sp_func_variable, b->variable)) {
            /* 3) process functions */
            replaced = sp_process_functions(r, replaced);
            apr_table_set(r->subprocess_env, b->variable, replaced);
          }
        }
      }
    }
  }
  return;
}

/**
 * Same as sp_match_var() but this method adds a request query parameter
 */
static void sp_match_query(request_rec *r, sp_std_entry_t *b, const char *val) {
  if(val) {
    ap_regmatch_t regm[AP_MAX_REG_MATCH];
    apr_size_t val_len = strlen(val);
    if((b->pattern && apr_strmatch(b->pattern, val, val_len)) ||
       (!b->pattern && !ap_regexec(b->preg, val, AP_MAX_REG_MATCH, regm, 0))) {
      char *replaced = NULL;
      char *query = apr_pstrdup(r->pool, b->variable);
      if(b->valueSet) {
        if(!b->pattern) {
          /* 1) respolve sub regex */
          replaced = ap_pregsub(r->pool, b->value, val, AP_MAX_REG_MATCH, regm);
        } else {
          /* literal string (no $1 .. $9) */
          replaced = apr_pstrdup(r->pool, b->value);
        }
      }
      if(replaced) {
        /* 2) sp_reslove_variable() resolves all ${xxx} variables */
        if(sp_reslove_variable(r, m_sp_var_pfx, 2, '}', &replaced,
                               &sp_func_variable, b->variable)) {
          /* 3) process functions */
          replaced = sp_process_functions(r, replaced);
        }
        query = apr_pstrcat(r->pool, query, "=", replaced, NULL);
      }
      if(r->args) {
        r->args = apr_pstrcat(r->pool, r->args, "&", query, NULL);
        r->unparsed_uri = apr_pstrcat(r->pool, r->unparsed_uri, "&", query, NULL);
      } else {
        r->args = query;
        r->unparsed_uri = apr_pstrcat(r->pool, r->unparsed_uri, "?", query, NULL);
      }
      if(r->parsed_uri.query) {
        r->parsed_uri.query = apr_pstrcat(r->pool, r->parsed_uri.query, "&", query, NULL);
      } else {
        r->parsed_uri.query = query;
      }
    }
  }
  return;
}

/**
 * Changes the request line query if the defined variable
 * has been set.
 *
 * @param r
 * @param vars Array of variable names
 */
static void sp_query(request_rec *r, apr_array_header_t *vars) {
  if(r->unparsed_uri == NULL) {
    return;
  }
  int i;
  char **entries = (char **)vars->elts;
  for(i = 0; i < vars->nelts; ++i) {
    char *variable = entries[i];
    const char *newquery = apr_table_get(r->subprocess_env, variable);
    if(newquery) {
      if(r->args) {
        // search the query within unparsed_uri
        char *q = r->unparsed_uri;
        while(q && q[0]) {
          q = strchr(q, '?');
          if(q) {
            if(q > r->unparsed_uri) {
              if(q[-1] != '\\') {
                q[0] = '\0';
              } else {
                // escaped, search next
                q++;
              }
            } else {
              q[0] = '\0';
            }
          }
        }
      }
      r->unparsed_uri = apr_pstrcat(r->pool, r->unparsed_uri, "?", newquery, NULL);
      r->args = apr_pstrdup(r->pool, newquery);
      r->parsed_uri.query = apr_pstrdup(r->pool, newquery);
    }
  }
  return;
}

/**
 * Hashes the configured variable and stores it as a request header
 */
static void sp_hash(request_rec *r, apr_array_header_t *header_hash) {
  int i;
  sp_hash_entry_t *entries = (sp_hash_entry_t *)header_hash->elts;
  for(i = 0; i < header_hash->nelts; ++i) {
    sp_hash_entry_t *b = &entries[i];
    const char *var = apr_table_get(r->subprocess_env, b->var);
    if(var) {
      char *md = ap_md5_binary(r->pool, (unsigned char *)var, strlen(var));
      char *hash = apr_pcalloc(r->pool, 64);
      char *d = hash;
      int c = 0;
      while(md[0]) {
        d[0] = md[0];
        d++;
        c++;
        md++;
        if(c == 4 && md[0]) {
          c=0;
          d[0] = ':';
          d++;
        }
      }
      d[0] = '\0';
      apr_table_set(r->headers_in, b->header, hash);
    }
  }
}

/**
 * Compares the variables defined by SetEnvIfCmpPlus and
 * adds the specified variable.
 *
 * @param r
 * @param cmps Rules to process
 */
static void sp_cmp(request_rec *r, apr_array_header_t * cmps) {
  int i;
  sp_cmp_entry_t *entries = (sp_cmp_entry_t *)cmps->elts;
  for(i = 0; i < cmps->nelts; ++i) {
    sp_cmp_entry_t *b = &entries[i];
    const char *leftStr = apr_table_get(r->subprocess_env, b->left);
    const char *rightStr = apr_table_get(r->subprocess_env, b->right);
    if(leftStr != NULL && rightStr != NULL) {
      int set = 0;
      if(sp_isnum(leftStr) && sp_isnum(rightStr)) {
        int left = atoi(leftStr);
        int right = atoi(rightStr);
        switch (b->cmp) {
        case SP_CMP_EQ:
          if(left == right) {
            set = 1;
          }
          break;
        case SP_CMP_NE:
          if(left != right) {
            set = 1;
          }
          break;
        case SP_CMP_GT:
          if(left > right) {
            set = 1;
          }
          break;
        case SP_CMP_LT:
          if(left < right) {
            set = 1;
          }
          break;
        }
      } else {
        int c = strcasecmp(leftStr, rightStr);
        switch (b->cmp) {
        case SP_CMP_EQ:
          if(c == 0) {
            set = 1;
          }
          break;
        case SP_CMP_NE:
          if(c != 0) {
            set = 1;
          }
          break;
        case SP_CMP_GT:
          if(c < 0) {
            set = 1;
          }
          break;
        case SP_CMP_LT:
          if(c > 0) {
            set = 1;
          }
          break;
        }
      }
      if(set) {
        if(b->variable[0] == '!') {
          apr_table_unset(r->subprocess_env, &b->variable[1]);
        } else {
          char *replaced = apr_pstrdup(r->pool, b->value);
          sp_reslove_variable(r, m_sp_var_pfx, 2, '}', &replaced, &sp_func_variable, NULL);
          apr_table_set(r->subprocess_env, b->variable, replaced);
        }
      }
    }
  }
}

/**
 * Determines the attriubte to match and calls sp_match_*() in order
 * to (un-)set the variable or query
 *
 * @param r
 * @pram headers HTTP headers (request or response)
 * @param conditionals Ruls to process
 * @param isQuery Sets a query parameter instead of an env variable
 * @return DECLINED
 */
static int sp_setenv(request_rec *r, apr_table_t *headers,
                     apr_array_header_t *conditionals, int isQuery) {
  int i;
  sp_std_entry_t *entries = (sp_std_entry_t *)conditionals->elts;
  void (*sp_match)(request_rec *, sp_std_entry_t *, const char *);
  if(isQuery) {
    sp_match = &sp_match_query;
  } else {
    sp_match = &sp_match_var;
  }

  for(i = 0; i < conditionals->nelts; ++i) {
    sp_std_entry_t *b = &entries[i];
    const char *val = NULL;
    
    switch (b->special_type) {
    case SP_SPECIAL_REMOTE_ADDR:
      val = SF_CONN_REMOTEIP(r);
      sp_match(r, b, val);
      break;
    case SP_SPECIAL_REMOTE_NET:
      val = sp_get_remote_net(r);
      sp_match(r, b, val);
      break;
    case SP_SPECIAL_SERVER_ADDR:
      val = r->connection->local_ip;
      sp_match(r, b, val);
      break;
    case SP_SPECIAL_SERVER_PORT:
      val = apr_psprintf(r->pool, "%d", r->connection->local_addr->port);
      sp_match(r, b, val);
      break;
    case SP_SPECIAL_REMOTE_HOST:
      val =  ap_get_remote_host(r->connection, r->per_dir_config, REMOTE_NAME, NULL);
      sp_match(r, b, val);
      break;
    case SP_SPECIAL_REQUEST_URI:
      val = r->uri;
      sp_match(r, b, val);
      break;
    case SP_SPECIAL_REQUEST_METHOD:
      val = r->method;
      sp_match(r, b, val);
      break;
    case SP_SPECIAL_REQUEST_PROTOCOL:
      val = r->protocol;
      sp_match(r, b, val);
      break;
    case SP_SPECIAL_REQUEST_QUERY:
      val = r->args;
      sp_match(r, b, val);
      break;
    case SP_SPECIAL_REQUEST_USER:
      val = r->user;
      sp_match(r, b, val);
      break;
    case SP_SPECIAL_RESPONSE_STATUS:
      val = apr_psprintf(r->pool, "%d", r->status);
      sp_match(r, b, val);
      break;
    case SP_SPECIAL_NOT:
      if (b->pnamereg) {
        /* Matching headers_in against a regex. Iterate through
         * the headers until we find a match or run out of
         * headers.
         */
        int j;
        const apr_array_header_t *arr = apr_table_elts(headers);
        const apr_table_entry_t *elts = (const apr_table_entry_t *)arr->elts;
        for(j = 0; j < arr->nelts; ++j) {
          if(!ap_regexec(b->pnamereg, elts[j].key, 0, NULL, 0)) {
            val = elts[j].val;
            sp_match(r, b, val);
          }
        }
      } else {
        /* Not matching against a regex */
        val = apr_table_get(headers, b->name);
        if(val == NULL) {
          val = apr_table_get(r->subprocess_env, b->name);
        }
        sp_match(r, b, val);
      }
    }
  }  
  return DECLINED;
}

/**
 * Evaluates a condition variable by looking it up within the
 * provided process variable table.
 *
 * @param subprocess_env Process environment variable table
 * @param conftition Variable name (optionally negotiated by a '!')
 *        to lookup
 * @return 1 If the variable is set or 0 if not
 */
static int sp_condition(apr_table_t *subprocess_env, const char *condition) {
  if(condition[0] == '!') {
    if(!apr_table_get(subprocess_env, &condition[1])) {
      return 1;
    }
  } else {
    if(apr_table_get(subprocess_env, condition)) {
      return 1;
    }
  }
  return 0;
}

/**
 * Modifies the request or response headers as configured by on of the
 * following directives:
 * - ChangeRequestHeaderPlus
 * - ChangeResponseHeaderPlus
 *
 * @param r
 * @param headers Either the in or out headers table
 * @param change_rules Configuration to apply
 */
static void sp_change_header(request_rec *r, apr_table_t *headers, apr_table_t *change_rules) {
  ap_regmatch_t regm[AP_MAX_REG_MATCH];
  apr_table_t *changed = apr_table_make(r->pool, 20); /* modified headers */
  apr_table_t *orig = apr_table_make(r->pool, 20);    /* unchanged headers */
  int i;
  apr_table_entry_t *header = (apr_table_entry_t *)apr_table_elts(headers)->elts;
  for(i = 0; i < apr_table_elts(headers)->nelts; i++) {
    int match = 0;
    int k;
    apr_table_entry_t *rule = (apr_table_entry_t *)apr_table_elts(change_rules)->elts;
    for(k = 0; k < apr_table_elts(change_rules)->nelts; k++) {
      sp_hdr_entry_t *b = (sp_hdr_entry_t *)rule[k].val;
      if(strcasecmp(b->name, header[i].key) == 0) {
        if(!ap_regexec(b->preg, header[i].val, AP_MAX_REG_MATCH, regm, 0)) {
          char *replaced = ap_pregsub(r->pool, b->value, header[i].val, AP_MAX_REG_MATCH, regm);
          sp_reslove_variable(r, m_sp_var_pfx, 2, '}', &replaced, &sp_func_variable, NULL);
          replaced = sp_process_functions(r, replaced);
          apr_table_addn(changed, header[i].key, replaced);
          match = 1;
          // content type of the http response is stored within the request record too
          if((strcasecmp("content-type", b->name) == 0) &&
             r->content_type &&
             (strcasecmp(r->content_type, header[i].val) == 0)){
            r->content_type = replaced;
          }
          break;
        }
      }
    }
    if(!match) {
      apr_table_addn(orig, header[i].key, header[i].val);
    }
  }
  apr_table_clear(headers);
  header = (apr_table_entry_t *)apr_table_elts(orig)->elts;
  for(i = 0; i < apr_table_elts(orig)->nelts; i++) {
    apr_table_addn(headers, header[i].key, header[i].val);
  }
  header = (apr_table_entry_t *)apr_table_elts(changed)->elts;
  for(i = 0; i < apr_table_elts(changed)->nelts; i++) {
    apr_table_addn(headers, header[i].key, header[i].val);
  }
}

/**
 * Removes the pattern matching the regular expression from
 * the header table.
 * Empty header is removed.
 *
 * @param r
 * @param header Headers to process
 * @param list Rules (patterns for header names)*
 */
static void sp_removepattern(request_rec *r, apr_table_t *headers, apr_array_header_t *list) {
  int i;
  sp_hdr_entry_t *entries = (sp_hdr_entry_t *)list->elts;
  ap_regmatch_t ma[AP_MAX_REG_MATCH];
  for(i = 0; i < list->nelts; ++i) {
    sp_hdr_entry_t *b = &entries[i];
    const char *header = apr_table_get(headers, b->name);
    if(header) {
      int changed = 0;
      while(!ap_regexec(b->preg, header, AP_MAX_REG_MATCH, ma, 0)) {
        int j = 1;
        const char *h = &header[ma[j].rm_eo];
        header = apr_pstrndup(r->pool, header, ma[j].rm_so);
        header = apr_pstrcat(r->pool, header, h, NULL);
        changed++;
      }
      if(changed) {
        if(header == NULL || header[0] == '\0') {
          apr_table_unset(headers, b->name);
        } else {
          apr_table_set(headers, b->name, header);
        }
      }
    }
  }
  return;
}

/**
 * Adds the specified output filters
 * @param r
 * @param filters
 */
static void sp_outfilter(request_rec *r, apr_array_header_t *filters) {
  int i;
  sp_filter_entry_t *entries = (sp_filter_entry_t *)filters->elts;
  apr_table_t *addedfilter = NULL; // ensure we add the filter only once
  for(i = 0; i < filters->nelts; ++i) {
    sp_filter_entry_t *entry = &entries[i];
    const char *var = entry->var;
    if(entry->notset) {
      if(apr_table_get(r->subprocess_env, var) == NULL) {
        if(addedfilter == NULL) {
          addedfilter = apr_table_make(r->pool, 2);
        }
        if(apr_table_get(addedfilter, entry->filtername) == NULL) {
          apr_table_add(addedfilter, entry->filtername, "");
          ap_add_output_filter(entry->filtername, NULL, r, r->connection);
        }
      }
    } else {
      if(apr_table_get(r->subprocess_env, var) != NULL) {
        if(addedfilter == NULL) {
          addedfilter = apr_table_make(r->pool, 2);
        }
        if(apr_table_get(addedfilter, entry->filtername) == NULL) {
          apr_table_add(addedfilter, entry->filtername, "");
          ap_add_output_filter(entry->filtername, NULL, r, r->connection);
        }
      }
    }
  }
}

/**
 * Changes the response status code if the configured
 * event is set.
 */
static void sp_status(request_rec *r, apr_array_header_t *vars) {
  int i;
  sp_status_entry_t *entries = (sp_status_entry_t *)vars->elts;
  for(i = 0; i < vars->nelts; ++i) {
    sp_status_entry_t *entry = &entries[i];
    const char *var = entry->var;
    if(apr_table_get(r->subprocess_env, var) != NULL) {
      r->status = entry->code;
    }
  }
}

/**
 * sets/unsets/adds headers to the provided headers table (which is either
 * the in or out headers table)
 *
 * note: Server and Date response header can't be set on locally (no
 *       mod_proxy) served HTTP request because basic_http_header()
 *       of http_filters.c writes direcly to the brigade removing 
 *       and Server/Date header set by mod_setenvifplus (or other)
 *       modules.
 *
 * @param r
 * @param headers Either the in or out headers to modify
 * @param list Rules to process (std_req_header, std_late_req_header
 *        or std_res_header)
 * @param isReponse Indicates if we are processing the reqest (0) or
 *         response (1) header fields.
 */
static int sp_header(request_rec *r, apr_table_t *headers, apr_array_header_t *list,
                     int isResponse) {
  apr_table_t *change_rules = NULL;
  int i;
  sp_hdr_entry_t *entries = (sp_hdr_entry_t *)list->elts;
  for(i = 0; i < list->nelts; ++i) {
    char *val;
    sp_hdr_entry_t *b = &entries[i];
    switch (b->action) {
    case SP_ACTION_SET:
      val = apr_pstrdup(r->pool, b->value);
      if(sp_reslove_variable(r, m_sp_var_pfx, 2, '}', &val, &sp_func_variable, NULL)) {
        if(b->condition) {
          if(sp_condition(r->subprocess_env, b->condition)) {
            apr_table_set(headers, b->name, val);
            // content type of the http response is stored within the request record too
            if((strcasecmp("content-type", b->name) == 0) && isResponse) {
              r->content_type = val;
            }
          }
        } else {
          apr_table_set(headers, b->name, val);
          // content type of the http response is stored within the request record too
          if((strcasecmp("content-type", b->name) == 0) && isResponse) {
            r->content_type = val;
          }
        }
      }
      break;
    case SP_ACTION_UNSET:
      if(b->condition) {
        if(sp_condition(r->subprocess_env, b->condition)) {
          apr_table_unset(headers, b->name);
        }
      } else {
        apr_table_unset(headers, b->name);
      }
      break;
    case SP_ACTION_ADD:
      val = apr_pstrdup(r->pool, b->value);
      if(sp_reslove_variable(r, m_sp_var_pfx, 2, '}', &val, &sp_func_variable, NULL)) {
        if(b->condition) {
          if(sp_condition(r->subprocess_env, b->condition)) {
            apr_table_add(headers, b->name, val);
            // content type of the http response is stored within the request record too
            if((strcasecmp("content-type", b->name) == 0) && isResponse) {
              r->content_type = val;
            }
          }
        } else {
          apr_table_add(headers, b->name, val);
          // content type of the http response is stored within the request record too
          if((strcasecmp("content-type", b->name) == 0) && isResponse) {
            r->content_type = val;
          }
        }
      }
      break;
    case SP_ACTION_CHANGE:
      if(!change_rules) {
        change_rules = apr_table_make(r->pool, 1);
      }
      apr_table_addn(change_rules, apr_psprintf(r->pool, "%d", i), (char *)b);
      break;
    }
  }
  if(change_rules) {
    sp_change_header(r, headers, change_rules);
  }
  return DECLINED;
}

/**
 * Extract the session cookie from the request.
 *
 * @param r
 * @param name Name of the cookie
 * @return Cookie or NULL if not found within the request headers
 */
static char *sp_get_remove_cookie(request_rec *r, const char *name) {
  const char *cookie_h = apr_table_get(r->headers_in, "Cookie");
  if(cookie_h) {
    char *cn = apr_pstrcat(r->pool, name, "=", NULL);
    char *pt = ap_strcasestr(cookie_h, cn);
    char *p = NULL;
    while(pt && !p) {
      // ensure we found the real cookie (and not an ending b64 str)
      if(pt == cookie_h) {
        // @beginning of the header
        p = pt;
        pt = NULL;
      } else {
        char pre = pt[-1];
        if(pre == ' ' ||
           pre == ';') {
          // @beginnin of a cookie
          p = pt;
          pt = NULL;
        } else {
          // found patter somewhere else
          pt++;
          pt = ap_strcasestr(pt, cn);
        }
      }
    }
    if(p) {
      char *value = NULL;
      char *clean = p;
      if(clean > cookie_h) {
        clean--;
        while(clean > cookie_h && clean[0] == ' ') {
          clean[0] = '\0';
          clean--;
        }
      }
      p[0] = '\0'; /* terminate the beginning of the cookie header */
      p = p + strlen(cn);
      value = ap_getword(r->pool, (const char **)&p, ';');
      while(p && (p[0] == ' ')) p++;
      /* skip a path, if there is any */
      if(p && (strncasecmp(p, "$path=", strlen("$path=")) == 0)) {
        ap_getword(r->pool, (const char **)&p, ';');
      }
      /* restore cookie header */
      if(p && p[0]) {
        if(cookie_h[0]) {
          cookie_h = apr_pstrcat(r->pool, cookie_h, " ", p, NULL);
        } else {
          cookie_h = apr_pstrcat(r->pool, p, NULL);
        }
      }
      if(strlen(cookie_h) == 0) {
        apr_table_unset(r->headers_in, "cookie");
      } else {
        if((strncasecmp(cookie_h, "$Version=", strlen("$Version=")) == 0) &&
           (strlen(cookie_h) <= strlen("$Version=X; "))) {
          /* nothing left */
          apr_table_unset(r->headers_in, "cookie");
        } else {
          apr_table_set(r->headers_in, "Cookie", cookie_h);
        }
      }
      return value;
    }
  }
  return NULL;
}

/**
 * Encrypts the specified Set-Cookie headers
 *
 * @param r 
 * @param key
 * @param cookie_names List of cookies to be encrypted
 * @param val Set-Cookie header value to process
 * @param header2keep Table with the modified Set-Cookie headers
 *        which shall replace all other Set-Cookie headers in the
 *        response.
 * @param modified Indicates, the the value has been modified or not
 */
static void sp_encSecCookie(request_rec *r, unsigned char *key,
                            apr_array_header_t *cookie_names,
                            const char *val,
                            apr_table_t *headers2keep, int *modified) {
  int i;
  char **names = (char **)cookie_names->elts;
  for(i = 0; i < cookie_names->nelts; ++i) {
    char *name = apr_pstrcat(r->pool, names[i], "=", NULL);
    int len = strlen(name);
    if(strncmp(val, name, len) == 0) {
      char *p = apr_pstrdup(r->pool, val);
      char *options = strchr(p, ';');
      char *v = strchr(p, '=');
      v++;
      if(options) {
        options[0] = '\0';
        options++;
      }
      v = sp_enc64(r, v, key);
      apr_table_add(headers2keep, "Set-Cookie",
                    apr_pstrcat(r->pool, name, v, options ? ";" : "", options, NULL));
      *modified = 1;
      return;
    }
  }
  // add without modification
  apr_table_add(headers2keep, "Set-Cookie", val);
}

/**
 * Processes the outgoing Set-Cookie header fields encrypting
 * the specified cookies.
 *
 * @param r
 * @param headers Table of the outgoing HTTP headers
 */
static void sp_setcookie(request_rec *r, apr_table_t *headers) {
  sp_config_t *sconf = ap_get_module_config(r->server->module_config, &setenvifplus_module);
  apr_array_header_t *cookie_names = sconf->cookie_names;
  if(cookie_names->nelts > 0) {
    const char *keyStr = apr_table_get(r->subprocess_env, SP_COOKIE_KEY);
    if(keyStr) {
      int i;
      int modified = 0;
      apr_table_t *headers2keep = apr_table_make(r->pool, 2);
      apr_table_entry_t *entry = (apr_table_entry_t *)apr_table_elts(headers)->elts;
      unsigned char key[EVP_MAX_KEY_LENGTH];
      EVP_BytesToKey(EVP_des_ede3_cbc(), EVP_sha1(), NULL, (unsigned char *)keyStr,
                     strlen(keyStr), 1, key, NULL);
      for(i = 0; i < apr_table_elts(headers)->nelts; i++) {
        if(strcasecmp(entry[i].key, "Set-Cookie") == 0) {
          sp_encSecCookie(r, key, cookie_names, entry[i].val, headers2keep, &modified);
        }
      }
      if(modified) {
        apr_table_unset(headers, "Set-Cookie");
        entry = (apr_table_entry_t *)apr_table_elts(headers2keep)->elts;
        for(i = 0; i < apr_table_elts(headers2keep)->nelts; i++) {
          apr_table_add(headers, entry[i].key, entry[i].val);
        }
      }
    }
  }  
}

/**
 * Processes the incomming HTTP Cookie header decrypting
 * the specified cookies.
 *
 * @param r
 * @param conf
 */
static void sp_cookie(request_rec *r, sp_config_t *conf) {
  apr_array_header_t *cookie_names = conf->cookie_names;
  const char *decCookies = NULL;
  if(cookie_names->nelts > 0) {
    const char *keyStr = apr_table_get(r->subprocess_env, SP_COOKIE_KEY);
    if(keyStr) {
      // rule and key available
      const char *cookie2 = NULL;
      int i;
      char **names = (char **)cookie_names->elts;
      unsigned char key[EVP_MAX_KEY_LENGTH];
      EVP_BytesToKey(EVP_des_ede3_cbc(), EVP_sha1(), NULL, (unsigned char *)keyStr,
                     strlen(keyStr), 1, key, NULL);
      for(i = 0; i < cookie_names->nelts; ++i) {
        char **name = &names[i];
        char *cookie = sp_get_remove_cookie(r, *name);
        if(cookie) {
          cookie = sp_dec64(r, cookie, key);
          if(cookie) {
            if(decCookies) {
              decCookies = apr_pstrcat(r->pool, decCookies, "; ", *name, "=", cookie, NULL);
            } else {
              decCookies = apr_pstrcat(r->pool, *name, "=", cookie, NULL);
            }
          } else {
            ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r,
                          SP_LOG_PFX(040)"Drop cookie %s.", *name);
          }
        }
        // TODO support cookie2 too
        // no cookie2 header support, drop the header if we found the configured cookie
        cookie2 = apr_table_get(r->headers_in, "Cookie2");
        if(cookie2) {
          char *search = apr_pstrcat(r->pool, *name, "=", NULL);
          if(strstr(cookie2, search) != NULL) {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                          SP_LOG_PFX(041)"Drop cookie2 header '%s.'", cookie2);
            apr_table_unset(r->headers_in, "Cookie2");
          }
        }
      }
    }    
  }
  if(decCookies) {
    const char *cookie = apr_table_get(r->headers_in, "Cookie");
    if(cookie) {
      int cookieLen = strlen(cookie);
      if(cookieLen > 1 && cookie[cookieLen-1] == ';') {
        cookie = apr_pstrcat(r->pool, cookie, " ", decCookies, NULL);
      } else {
        cookie = apr_pstrcat(r->pool, cookie, "; ", decCookies, NULL);
      }
    } else {
      cookie = decCookies;
    }
    apr_table_set(r->headers_in, "Cookie", cookie);
  }
}

/************************************************************************
 * handlers
 ***********************************************************************/

static int sp_fixup(request_rec *r) {
  sp_config_t *sconf = ap_get_module_config(r->server->module_config, &setenvifplus_module);
  sp_config_t *conf = ap_get_module_config(r->per_dir_config, &setenvifplus_module);
  if(conf) {
    /* SetEnvIfPlus late */
    sp_setenv(r, r->headers_in, conf->std_late_conditionals, 0);
    /* SetEnvIfCmpPlus late */
    sp_cmp(r, conf->cmp_late);
    /*  RequestHeaderPlus late */
    sp_header(r, r->headers_in, conf->std_late_req_header, 0);
  }
  if(sconf) {
    sp_cookie(r, sconf);
  }
  return DECLINED;
}

static int sp_user_id(request_rec *r) {
  sp_config_t *conf = ap_get_module_config(r->per_dir_config, &setenvifplus_module);
  if(conf) {
    ap_regmatch_t regm[AP_MAX_REG_MATCH];
    apr_array_header_t *c = conf->user;
    int i;
    sp_usr_entry_t *entries = (sp_usr_entry_t *)c->elts;
    for(i = 0; i < c->nelts; ++i) {
      sp_usr_entry_t *b = &entries[i];
      const char *hdr = apr_table_get(r->headers_in, b->header);
      if(hdr) {
        if(!ap_regexec(b->preg, hdr, AP_MAX_REG_MATCH, regm, 0)) {
          char *replaced = ap_pregsub(r->pool, b->value, hdr, AP_MAX_REG_MATCH, regm);
          if(replaced) {
            r->user = replaced;
            r->ap_auth_type = (char*)apr_pstrdup(r->pool, "delegated");
            return OK;
          }
        }
      }
    }
  }
  return DECLINED;
}

static int sp_header_parser(request_rec *r) {
  sp_config_t *conf = ap_get_module_config(r->per_dir_config, &setenvifplus_module);
  if(conf) {
    /* SetEnvIfPlus */
    sp_setenv(r, r->headers_in, conf->std_conditionals, 0);
    /*  RequestHeaderPlus ChangeRequestHeaderPlus */
    sp_header(r, r->headers_in, conf->std_req_header, 0);
    /* ChangeQueryIfPlus */
    sp_query(r, conf->std_req_changequery);
    /* SetQueryIfPlus */
    sp_setenv(r, r->headers_in, conf->std_req_query, 1);
    /* SetEnvIfCmpPlus */
    sp_cmp(r, conf->cmp);

    sp_removepattern(r, r->headers_in, conf->std_removepattern_req_header);
  }
  return DECLINED;
}

static int sp_post_read_request(request_rec *r) {
  sp_config_t *conf = ap_get_module_config(r->server->module_config, &setenvifplus_module);
  if(conf) {
    /* SetEnvIfPlus */
    sp_setenv(r, r->headers_in, conf->std_conditionals, 0);
    /*  RequestHeaderPlus ChangeRequestHeaderPlus */
    sp_header(r, r->headers_in, conf->std_req_header, 0);
    /* ChangeQueryIfPlus */
    sp_query(r, conf->std_req_changequery);
    /* SetQueryIfPlus */
    sp_setenv(r, r->headers_in, conf->std_req_query, 1);
    /* SetEnvIfCmpPlus */
    sp_cmp(r, conf->cmp);
    /* SetHashHeaderPlus */
    sp_hash(r, conf->header_hash);

  }
  return DECLINED;
}

#ifdef SP_INTERNAL_TEST
/** test handler printing all variables */
static int sp_handler(request_rec * r) {
  if(strcmp(r->handler, "setenvifplus-test")) {
    return DECLINED;
  } else {
    int i;
    apr_table_entry_t *entry = (apr_table_entry_t *)apr_table_elts(r->subprocess_env)->elts;
    ap_set_content_type(r, "text/plain");
    for(i = 0; i < apr_table_elts(r->subprocess_env)->nelts; i++) {
      ap_rprintf(r, "var: <%s><%s>\n", entry[i].key, entry[i].val);
    }
    entry = (apr_table_entry_t *)apr_table_elts(r->headers_in)->elts;
    ap_set_content_type(r, "text/plain");
    apr_table_set(r->headers_out, "X-SetEnvIfPlus",
                  apr_psprintf(r->pool, "testhandler=%s", g_revision));
    for(i = 0; i < apr_table_elts(r->headers_in)->nelts; i++) {
      ap_rprintf(r, "hdr: <%s><%s>\n", entry[i].key, entry[i].val);
    }

    ap_rprintf(r, "REMOTE_ADDR: <%s>\n", SF_CONN_REMOTEIP(r));
    ap_rprintf(r, "REMOTE_HOST: <%s>\n", ap_get_remote_host(r->connection, 
                                                            r->per_dir_config, REMOTE_NAME, NULL));
    ap_rprintf(r, "REMOTE_NET: <%s>\n", sp_get_remote_net(r));

    ap_rprintf(r, "SERVER_ADDR: <%s>\n", r->connection->local_ip);
    ap_rprintf(r, "SERVER_PORT: <%d>\n", r->connection->local_addr->port);
  }
  return OK;
}
#endif

/** finalize configuration */
static int sp_post_config(apr_pool_t *pconf, apr_pool_t *plog, apr_pool_t *ptemp,
                            server_rec *bs) {
  ap_add_version_component(pconf, apr_psprintf(pconf, "mod_setenvifplus/%s", g_revision));
#ifdef SP_INTERNAL_TEST
  fprintf(stdout, "\033[1mmod_setenvifplus TEST BINARY, NOT FOR PRODUCTIVE USE\033[0m\n");
  fflush(stdout);
#endif
  return DECLINED;
}

/**
 * Response processing
 * @param r
 * @paramheaders HTTP response headers
 */
static void sp_out(request_rec *r, apr_table_t *headers) {
  sp_config_t *dconf = ap_get_module_config(r->per_dir_config, &setenvifplus_module);
  sp_config_t *sconf = ap_get_module_config(r->server->module_config, &setenvifplus_module);

  if(sconf) {
    sp_header(r, headers, sconf->std_res_header, 1);
    sp_status(r, sconf->status_var);
    sp_outfilter(r, sconf->outfilter);
  }
  if(dconf) {
    sp_setenv(r, headers, dconf->std_res_conditionals, 0);
    sp_header(r, headers, dconf->std_res_header, 1);
    sp_status(r, dconf->status_var);
    sp_outfilter(r, dconf->outfilter);
  }
  return;
}

static apr_status_t sp_out_err_filter(ap_filter_t *f, apr_bucket_brigade *bb) {
  request_rec *r = f->r;
  sp_out(r, r->err_headers_out);
  sp_setcookie(r, r->err_headers_out);
  ap_remove_output_filter(f);
  return ap_pass_brigade(f->next, bb); 
}

static apr_status_t sp_out_filter(ap_filter_t *f, apr_bucket_brigade *bb) {
  request_rec *r = f->r;
  sp_out(r, r->headers_out);
  sp_setcookie(r, r->headers_out);
  ap_remove_output_filter(f);
  return ap_pass_brigade(f->next, bb); 
}

/**
 * insert response filter
 */
static void sp_insert_filter(request_rec *r) {
  ap_add_output_filter("sp_out_filter", NULL, r, r->connection);
}

static void sp_insert_err_filter(request_rec *r) {
  ap_add_output_filter("sp_out_err_filter", NULL, r, r->connection);
}

/************************************************************************
 * directiv handlers 
 ***********************************************************************/
static void *sp_srv_config_create(apr_pool_t *p, server_rec *s) {
  sp_config_t *sconf = apr_pcalloc(p, sizeof(sp_config_t));
  sconf->path = NULL;
  sconf->std_conditionals = apr_array_make(p, 20, sizeof(sp_std_entry_t));
  sconf->std_late_conditionals = NULL;
  sconf->std_req_query = apr_array_make(p, 20, sizeof(sp_std_entry_t));
  sconf->std_req_changequery = apr_array_make(p, 20, sizeof(char *));
  sconf->std_res_conditionals = NULL;
  sconf->std_req_header = apr_array_make(p, 20, sizeof(sp_hdr_entry_t));
  sconf->std_late_req_header = NULL;
  sconf->std_res_header = apr_array_make(p, 20, sizeof(sp_hdr_entry_t));
  sconf->cmp = apr_array_make(p, 2, sizeof(sp_cmp_entry_t));
  sconf->user = NULL;
  sconf->status_var = apr_array_make(p, 2, sizeof(sp_status_entry_t));
  sconf->header_hash = apr_array_make(p, 2, sizeof(sp_hash_entry_t));
  sconf->outfilter = apr_array_make(p, 2, sizeof(sp_filter_entry_t));
  sconf->cookie_names = apr_array_make(p, 2, sizeof(char *));
  return sconf;
}

static void *sp_srv_config_merge(apr_pool_t *p, void *basev, void *addv) {
  sp_config_t *sconf = apr_pcalloc(p, sizeof(sp_config_t));
  sp_config_t *base = basev;
  sp_config_t *add = addv;
  sconf->path = NULL;
  sconf->std_conditionals = apr_array_append(p, base->std_conditionals,
                                             add->std_conditionals);
  sconf->std_late_conditionals = NULL;
  sconf->std_req_query = apr_array_append(p, base->std_req_query,
                                             add->std_req_query);
  sconf->std_req_changequery = apr_array_append(p, base->std_req_changequery,
                                                add->std_req_changequery);
  sconf->std_res_conditionals = NULL;
  sconf->std_req_header = apr_array_append(p, base->std_req_header,
                                           add->std_req_header);
  sconf->std_late_req_header = NULL;
  sconf->std_res_header = apr_array_append(p, base->std_res_header,
                                           add->std_res_header);
  sconf->cmp = apr_array_append(p, base->cmp, add->cmp);
  sconf->status_var = apr_array_append(p, base->status_var, add->status_var);
  sconf->header_hash = apr_array_append(p, base->header_hash, add->header_hash);
  sconf->outfilter = apr_array_append(p, base->outfilter, add->outfilter);
  sconf->cookie_names = apr_array_append(p, base->cookie_names, add->cookie_names);
  return sconf;
}

static void *sp_dir_config_create(apr_pool_t *p, char *d) {
  sp_config_t *dconf = apr_pcalloc(p, sizeof(sp_config_t));
  dconf->path = d;
  dconf->std_conditionals = apr_array_make(p, 20, sizeof(sp_std_entry_t));
  dconf->std_late_conditionals = apr_array_make(p, 20, sizeof(sp_std_entry_t));
  dconf->std_req_query = apr_array_make(p, 20, sizeof(sp_std_entry_t));
  dconf->std_req_changequery = apr_array_make(p, 20, sizeof(char *));
  dconf->std_res_conditionals = apr_array_make(p, 20, sizeof(sp_std_entry_t));
  dconf->std_req_header = apr_array_make(p, 20, sizeof(sp_hdr_entry_t));
  dconf->std_late_req_header = apr_array_make(p, 20, sizeof(sp_hdr_entry_t));
  dconf->std_res_header = apr_array_make(p, 20, sizeof(sp_hdr_entry_t));
  dconf->cmp = apr_array_make(p, 2, sizeof(sp_cmp_entry_t));
  dconf->cmp_late = apr_array_make(p, 2, sizeof(sp_cmp_entry_t));
  dconf->user = apr_array_make(p, 2, sizeof(sp_usr_entry_t));
  dconf->std_removepattern_req_header = apr_array_make(p, 20, sizeof(sp_hdr_entry_t));
  dconf->status_var = apr_array_make(p, 2, sizeof(sp_status_entry_t));
  dconf->outfilter = apr_array_make(p, 2, sizeof(sp_filter_entry_t));
  return dconf;
}

static void *sp_dir_config_merge(apr_pool_t *p, void *basev, void *addv) {
  sp_config_t *dconf = apr_pcalloc(p, sizeof(sp_config_t));
  sp_config_t *base = basev;
  sp_config_t *add = addv;
  dconf->path = add->path;
  dconf->std_conditionals = apr_array_append(p, base->std_conditionals,
                                             add->std_conditionals);
  dconf->std_late_conditionals = apr_array_append(p, base->std_late_conditionals,
                                                  add->std_late_conditionals);
  dconf->std_req_query = apr_array_append(p, base->std_req_query,
                                          add->std_req_query);
  dconf->std_req_changequery = apr_array_append(p, base->std_req_changequery,
                                                add->std_req_changequery);
  dconf->std_res_conditionals = apr_array_append(p, base->std_res_conditionals,
                                                 add->std_res_conditionals);
  dconf->std_req_header = apr_array_append(p, base->std_req_header,
                                           add->std_req_header);
  dconf->std_late_req_header = apr_array_append(p, base->std_late_req_header,
                                                add->std_late_req_header);
  dconf->std_res_header = apr_array_append(p, base->std_res_header,
                                           add->std_res_header);
  dconf->cmp = apr_array_append(p, base->cmp, add->cmp);
  dconf->cmp_late = apr_array_append(p, base->cmp_late, add->cmp_late);
  dconf->user = apr_array_append(p, base->user, add->user);
  dconf->status_var = apr_array_append(p, base->status_var, add->status_var);
  dconf->outfilter = apr_array_append(p, base->outfilter, add->outfilter);
  dconf->std_removepattern_req_header = apr_array_append(p, base->std_removepattern_req_header,
                                           add->std_removepattern_req_header);
  return dconf;
}

const char *sp_header_cmd(cmd_parms *cmd, int argc, char *const argv[],
                          sp_hdr_entry_t *new) {
  const char *action;
  if(argc < 2) {
    /* always require action and header name */
    return apr_psprintf(cmd->pool, "%s: takes at least two arguments",
                        cmd->directive->directive);
  }
  action = argv[0];
  new->name = apr_pstrdup(cmd->pool, argv[1]);
  new->value = apr_pstrdup(cmd->pool, "");
  new->condition = NULL;
  if(strcasecmp(action, "set") == 0) {
    new->action = SP_ACTION_SET;
    if(argc < 3) {
      return apr_psprintf(cmd->pool, "%s set: takes at least three arguments",
                          cmd->directive->directive);
    }
    new->value = apr_pstrdup(cmd->pool, argv[2]);
    if(argc > 3) {
      char *c = argv[3];
      if(strncasecmp(c, "env=", strlen("env=")) == 0) {
        new->condition = apr_pstrdup(cmd->pool, &c[strlen("env=")]);
      } else {
        return apr_psprintf(cmd->pool, "%s set: expects 'env=' argument",
                            cmd->directive->directive);
      }
    }
  } else if(strcasecmp(action, "add") == 0) {
    new->action = SP_ACTION_ADD;
    if(argc < 3) {
      return apr_psprintf(cmd->pool, "%s add: takes at least three arguments",
                          cmd->directive->directive);
    }
    new->value = apr_pstrdup(cmd->pool, argv[2]);
    if(argc > 3) {
      char *c = argv[3];
      if(strncasecmp(c, "env=", strlen("env=")) == 0) {
        new->condition = apr_pstrdup(cmd->pool, &c[strlen("env=")]);
      } else {
        return apr_psprintf(cmd->pool, "%s add: expects 'env=' argument",
                            cmd->directive->directive);
      }
    }
  } else if(strcasecmp(action, "unset") == 0) {
    new->action = SP_ACTION_UNSET;
    if(argc > 2) {
      char *c = argv[2];
      if(strncasecmp(c, "env=", strlen("env=")) == 0) {
        new->condition = apr_pstrdup(cmd->pool, &c[strlen("env=")]);
      } else {
        return apr_psprintf(cmd->pool, "%s unset: expects 'env=' argument",
                            cmd->directive->directive);
      }
    }
  } else {
    return apr_psprintf(cmd->pool, "%s: unkown action",
                        cmd->directive->directive);
  }
  return NULL;
}

const char *sp_changeheader(cmd_parms *cmd, const char *header,
                            const char *regex, const char *value,
                            sp_hdr_entry_t *new) {
  new->action = SP_ACTION_CHANGE;
  new->preg = ap_pregcomp(cmd->pool, regex, (AP_REG_EXTENDED | AP_REG_ICASE));
  if(new->preg == NULL) {
    return apr_psprintf(cmd->pool, "%s: regex pattern '%s' could not be compiled",
                        cmd->directive->directive, regex);
  }
  new->name = apr_pstrdup(cmd->pool, header);
  new->value = apr_pstrdup(cmd->pool, value);
  return NULL;
}

const char *sp_changereqheader_cmd(cmd_parms *cmd, void *dcfg, const char *header,
                                   const char *regex, const char *value) {
  sp_hdr_entry_t *new;
  sp_config_t *conf;
  if(cmd->path) {
    conf = dcfg;
  } else {
    conf = ap_get_module_config(cmd->server->module_config, &setenvifplus_module);
  }
  new = apr_array_push(conf->std_req_header);
  return sp_changeheader(cmd, header, regex, value, new);
}

const char *sp_removepattern_cmd(cmd_parms *cmd, void *dcfg, const char *header,
                                 const char *regex) {
  sp_hdr_entry_t *new;
  sp_config_t *conf;
  char *pattern = apr_psprintf(cmd->pool, ".*(%s).*", regex);
  if(cmd->path) {
    conf = dcfg;
  } else {
    conf = ap_get_module_config(cmd->server->module_config, &setenvifplus_module);
  }
  new = apr_array_push(conf->std_removepattern_req_header);
  new->action = SP_ACTION_CHANGE;
  new->preg = ap_pregcomp(cmd->pool, pattern, (AP_REG_EXTENDED | AP_REG_ICASE));
  if(new->preg == NULL) {
    return apr_psprintf(cmd->pool, "%s: regex pattern '%s' could not be compiled",
                        cmd->directive->directive, pattern);
  }
  new->name = apr_pstrdup(cmd->pool, header);
  new->value = pattern;
  return NULL;
}

const char *sp_changeresheader_cmd(cmd_parms *cmd, void *dcfg, const char *header,
                                   const char *regex, const char *value) {
  sp_hdr_entry_t *new;
  sp_config_t *conf;
  if(cmd->path) {
    conf = dcfg;
  } else {
    conf = ap_get_module_config(cmd->server->module_config, &setenvifplus_module);
  }
  new = apr_array_push(conf->std_res_header);
  return sp_changeheader(cmd, header, regex, value, new);
}

const char *sp_cmp_cmd(cmd_parms *cmd, void *dcfg, int argc, char *const argv[]) {
  sp_cmp_entry_t *new;
  char *del;
  int late = 0;
  sp_config_t *conf;
  if(cmd->path) {
    conf = dcfg;
    if(argc == 5) {
      if(strcasecmp(argv[4], "late") == 0) {
        late = 1;
      } else {
        return apr_psprintf(cmd->pool, "%s: unknown parameter '%s'",
                            cmd->directive->directive, argv[4]);
      }
    } else if(argc != 4) {
      return apr_psprintf(cmd->pool, "%s: requires 4 or 5 arguments",
                          cmd->directive->directive);
    }
  } else {
    conf = ap_get_module_config(cmd->server->module_config, &setenvifplus_module);
    if(argc != 4) {
      return apr_psprintf(cmd->pool, "%s: requires 4 arguments",
                          cmd->directive->directive);
    }
  }
  if(late) {
    new = apr_array_push(conf->cmp_late);
  } else {
    new = apr_array_push(conf->cmp);
  }
  new->left = apr_pstrdup(cmd->pool, argv[0]);
  if(strcasecmp(argv[1], "eq") == 0) {
    new->cmp = SP_CMP_EQ;
  } else if(strcasecmp(argv[1], "ne") == 0) {
    new->cmp = SP_CMP_NE;
  } else if(strcasecmp(argv[1], "lt") == 0) {
    new->cmp = SP_CMP_LT;
  } else if(strcasecmp(argv[1], "gt") == 0) {
    new->cmp = SP_CMP_GT;
  } else {
    return apr_psprintf(cmd->pool, "%s: invalid operator '%s",
                        cmd->directive->directive, argv[1]);
  }
  new->right = apr_pstrdup(cmd->pool, argv[2]);
  new->variable = apr_pstrdup(cmd->pool, argv[3]);
  del = strchr(new->variable, '=');
  if(del) {
    new->value = &del[1];
    del[0] = '\0';
  } else {
    new->value = apr_pstrdup(cmd->pool, "");
  }
  return NULL;
}

/**
 * RequestHeaderPlus 
 */
const char *sp_reqheader_cmd(cmd_parms *cmd, void *dcfg, int argc, char *const argv[]) {
  sp_hdr_entry_t *new;
  sp_config_t *conf;
  if(cmd->path) {
    conf = dcfg;
  } else {
    conf = ap_get_module_config(cmd->server->module_config, &setenvifplus_module);
  }
  if(strcmp(argv[argc-1], "late") == 0) {
    if(!cmd->path) {
      return apr_psprintf(cmd->pool, "%s: 'late' option not allowed here",
                          cmd->directive->directive);
    }
    new = apr_array_push(conf->std_late_req_header);
    argc--;
  } else {
    new = apr_array_push(conf->std_req_header);
  }
  return sp_header_cmd(cmd, argc, argv, new);
}

const char *sp_resheader_cmd(cmd_parms *cmd, void *dcfg, int argc, char *const argv[]) {
  sp_hdr_entry_t *new;
  sp_config_t *conf;
  if(cmd->path) {
    conf = dcfg;
  } else {
    conf = ap_get_module_config(cmd->server->module_config, &setenvifplus_module);
  }
  new = apr_array_push(conf->std_res_header);
  return sp_header_cmd(cmd, argc, argv, new);
}

const char *sp_std_core_cmd(cmd_parms *cmd, apr_array_header_t *conditionals,
                            const char *name, const char *regex, const char *var,
                            int icase, int late) {
  const char *val;
  const char *simple_pattern;
  sp_std_entry_t *new;

  /* create a new entry */
  new = apr_array_push(conditionals);
  new->name = name;
  new->regex = regex;
  new->icase = icase;
  if((simple_pattern = sp_non_regex_pattern(cmd->pool, regex))) {
    new->pattern = apr_strmatch_precompile(cmd->pool, simple_pattern, !icase);
    if(new->pattern == NULL) {
      return apr_psprintf(cmd->pool, "%s: strmatch pattern '%s' could not be compiled",
                            cmd->directive->directive, simple_pattern);
    }
    new->preg = NULL;
  } else {
    new->preg = ap_pregcomp(cmd->pool, regex, (AP_REG_EXTENDED | (icase ? AP_REG_ICASE : 0)));
    if(new->preg == NULL) {
      return apr_psprintf(cmd->pool, "%s: regex pattern '%s' could not be compiled",
                          cmd->directive->directive, regex);
    }
    new->pattern = NULL;
  }
  if(!strcasecmp(name, "remote_addr")) {
    new->special_type = SP_SPECIAL_REMOTE_ADDR;
  } else if(!strcasecmp(name, "remote_net")) {
    new->special_type = SP_SPECIAL_REMOTE_NET;
  } else if(!strcasecmp(name, "remote_host")) {
    new->special_type = SP_SPECIAL_REMOTE_HOST;
  } else if(!strcasecmp(name, "request_uri")) {
    new->special_type = SP_SPECIAL_REQUEST_URI;
  } else if(!strcasecmp(name, "request_path")) {
    new->special_type = SP_SPECIAL_REQUEST_URI;
  } else if(!strcasecmp(name, "request_method")) {
    new->special_type = SP_SPECIAL_REQUEST_METHOD;
  } else if(!strcasecmp(name, "request_protocol")) {
    new->special_type = SP_SPECIAL_REQUEST_PROTOCOL;
  } else if(!strcasecmp(name, "server_addr")) {
    new->special_type = SP_SPECIAL_SERVER_ADDR;
  } else if(!strcasecmp(name, "server_port")) {
    new->special_type = SP_SPECIAL_SERVER_PORT;
  } else if(!strcasecmp(name, "request_query")) {
    new->special_type =  SP_SPECIAL_REQUEST_QUERY;
  } else if(!strcasecmp(name, "response_status")) {
    /* TODO: must be used for ResponseSetEnfIfPlus only */ 
    new->special_type =  SP_SPECIAL_RESPONSE_STATUS;
  } else if(!strcasecmp(name, "request_user")) {
    if(!late) {
      return apr_psprintf(cmd->pool, "%s: user match works only"
                          " when using the 'late' handler",
                          cmd->directive->directive);
    }
    new->special_type = SP_SPECIAL_REQUEST_USER;
  } else {
    new->special_type = SP_SPECIAL_NOT;
    /* Handle name as a regular expression.
     * If name a simple header string, identify as such
     * (new->pnamereg = NULL) to avoid the overhead of searching
     * through headers_in for a regex match.
     */
    if(sp_is_header_regex(cmd->pool, name)) {
      new->pnamereg = ap_pregcomp(cmd->pool, name,
                                  (AP_REG_EXTENDED | AP_REG_NOSUB
                                   | (icase ? AP_REG_ICASE : 0)));
      if (new->pnamereg == NULL)
      return apr_psprintf(cmd->pool, "%s: header name '%s' could not be compiled",
                          cmd->directive->directive, name);
    } else {
      new->pnamereg = NULL;
    }
  }
  val = strchr(var, '=');
  if(val) {
    new->variable = apr_pstrndup(cmd->pool, var, val-var);
    val++;
    new->value = apr_pstrdup(cmd->pool, val);
    new->valueSet = 1;
  } else {
    new->variable = apr_pstrdup(cmd->pool, var);
    new->value = apr_pstrdup(cmd->pool, "1");
    new->valueSet = 0;
  }
  return NULL;
}

/**
 * SetEnvIfPlus
 * syntax: 
 * <attribute> <regex> [!]<env-variable>[=<value>] [late]
 */
const char *sp_std_cmd(cmd_parms *cmd, void *dcfg, int argc, char *const argv[]) {
  int icase = cmd->info == SP_ICASE_MAGIC;
  sp_config_t *conf;
  apr_array_header_t *conditionals;
  int late = 0;
  const char *name;
  const char *regex;
  const char *var;
  if(argc < 3) {
    return apr_psprintf(cmd->pool, "%s: takes at least three arguments",
                        cmd->directive->directive);
  }
  name = argv[0];
  regex = argv[1];
  var = argv[2];
  /* determine scope of the directive */
  if(cmd->path) {
    if(argc > 4) {
      return apr_psprintf(cmd->pool, "%s: takes not more than four arguments",
                          cmd->directive->directive);
    }
    conf = dcfg;
    conditionals = conf->std_conditionals;
    if(argc == 4) {
      if(strcasecmp(argv[3], "late") == 0) {
        late = 1;
        conditionals = conf->std_late_conditionals;
      } else {
        return apr_psprintf(cmd->pool, "%s: unknown parameter '%s'",
                            cmd->directive->directive, argv[3]);
      }
    }
  } else {
    if(argc > 3) {
      return apr_psprintf(cmd->pool, "%s: takes not more than 3 arguments"
                          " when used at server level",
                          cmd->directive->directive);
    }
    conf = ap_get_module_config(cmd->server->module_config, &setenvifplus_module);
    conditionals = conf->std_conditionals;
  }

  return sp_std_core_cmd(cmd, conditionals, name, regex, var, icase, late);
}

/**
 * SetQueryIfPlus
 * syntax: 
 * <attribute> <regex> <query>[=<value>]
 */
const char *sp_std_req_query_cmd(cmd_parms *cmd, void *dcfg, const char *attribute,
                                 const char *regex, const char *query) {
  int icase = cmd->info == SP_ICASE_MAGIC;
  sp_config_t *conf;
  apr_array_header_t *req_query;
  /* determine scope of the directive */
  if(cmd->path) {
    conf = dcfg;
    req_query = conf->std_req_query;
  } else {
    conf = ap_get_module_config(cmd->server->module_config, &setenvifplus_module);
    req_query = conf->std_req_query;
  }

  return sp_std_core_cmd(cmd, req_query, attribute, regex, query, icase, 0);
}

const char *sp_std_req_changequery_cmd(cmd_parms *cmd, void *dcfg, const char *var) {
  sp_config_t *conf;
  apr_array_header_t *req_query;
  char **variable;
  /* determine scope of the directive */
  if(cmd->path) {
    conf = dcfg;
    req_query = conf->std_req_changequery;
  } else {
    conf = ap_get_module_config(cmd->server->module_config, &setenvifplus_module);
    req_query = conf->std_req_changequery;
  }
  variable = apr_array_push(req_query);
  *variable = apr_pstrdup(cmd->pool, var);
  return NULL;
}

/**
 * ResponseSetEnvIfPlus
 * syntax: 
 * <attribute> <regex> [!]<env-variable>[=<value>]
 */
const char *sp_res_std_cmd(cmd_parms *cmd, void *dcfg, int argc, char *const argv[]) {
  int icase = cmd->info == SP_ICASE_MAGIC;
  sp_config_t *conf = dcfg;
  int late = 0;
  const char *name;
  const char *regex;
  const char *var;
  if(argc != 3) {
    return apr_psprintf(cmd->pool, "%s: takes three arguments",
                        cmd->directive->directive);
  }
  name = argv[0];
  regex = argv[1];
  var = argv[2];
  return sp_std_core_cmd(cmd, conf->std_res_conditionals, name, regex, var, icase, late);
}

const char *sp_setuser_cmd(cmd_parms *cmd, void *dcfg, const char *header,
                           const char *regex, const char *value) {
  sp_usr_entry_t *new;
  sp_config_t *conf;
  if(cmd->path) {
    conf = dcfg;
  } else {
    conf = ap_get_module_config(cmd->server->module_config, &setenvifplus_module);
  }
  new = apr_array_push(conf->user);
  new->header = apr_pstrdup(cmd->pool, header);
  new->preg = ap_pregcomp(cmd->pool, regex, AP_REG_EXTENDED);
  if(new->preg == NULL) {
    return apr_psprintf(cmd->pool, "%s: regex pattern '%s' could not be compiled",
                        cmd->directive->directive, regex);
  }
  new->value = apr_pstrdup(cmd->pool, value);
  return NULL;
}

const char *sp_status_cmd(cmd_parms *cmd, void *dcfg, const char *status,
                           const char *var) {
  sp_status_entry_t *new;
  sp_config_t *conf;
  int idx500 = ap_index_of_response(HTTP_INTERNAL_SERVER_ERROR);
  if(cmd->path) {
    conf = dcfg;
  } else {
    conf = ap_get_module_config(cmd->server->module_config, &setenvifplus_module);
  }
  new = apr_array_push(conf->status_var);
  new->code = atoi(status);
  if(new->code == 0) {
    return apr_psprintf(cmd->pool, "%s: invalid status code",
                        cmd->directive->directive);
  }
  if(new->code != 500) {
    if(ap_index_of_response(new->code) == idx500) {
      return apr_psprintf(cmd->pool, "%s: unsupported HTTP response code", 
                          cmd->directive->directive);
    }
  }
  new->var = var;
  return NULL;
}


const char *sp_var_pfx_cmd(cmd_parms *cmd, void *dcfg, const char *pfx) {
  const char *err = ap_check_cmd_context(cmd, GLOBAL_ONLY);
  if (err != NULL) {
    return err;
  }
  if(strcmp(pfx, "$") == 0) {
    m_sp_var_pfx = apr_pstrdup(cmd->pool, "${");
  } else if(strcmp(pfx, "%") == 0) {
    m_sp_var_pfx = apr_pstrdup(cmd->pool, "%{");
  } else {
    return apr_psprintf(cmd->pool, "%s: unsupported variable prefix (use either %% or $)", 
                        cmd->directive->directive);
  }
  return NULL;
}

const char *sp_outfilter_cmd(cmd_parms *cmd, void *dcfg, const char *filter,
                             const char *var) {
  sp_filter_entry_t *new;
  sp_config_t *conf;
  if(cmd->path) {
    conf = dcfg;
  } else {
    conf = ap_get_module_config(cmd->server->module_config, &setenvifplus_module);
  }
  new = apr_array_push(conf->outfilter);
  new->filtername = filter;
  new->var = var;
  new->notset = 0;
  if(new->var[0] == '!') {
    new->var = &new->var[1];
    new->notset = 1;
  }
  return NULL;
}

const char *sp_cookie_cmd(cmd_parms *cmd, void *dcfg, const char *name) {
  char **new;
  sp_config_t *conf = ap_get_module_config(cmd->server->module_config, &setenvifplus_module);
  new = apr_array_push(conf->cookie_names);
  *new = apr_pstrdup(cmd->pool, name);
  return NULL;
}

const char *sp_hash_cmd(cmd_parms *cmd, void *dcfg, const char *header, const char *var) {
  sp_config_t *conf = ap_get_module_config(cmd->server->module_config, &setenvifplus_module);
  sp_hash_entry_t *new = apr_array_push(conf->header_hash);
  new->header = apr_pstrdup(cmd->pool, header);
  new->var = apr_pstrdup(cmd->pool, var);
  return NULL;
}

static const command_rec sp_config_cmds[] = {
  AP_INIT_TAKE_ARGV("SetEnvIfPlus", sp_std_cmd, NULL,
                    RSRC_CONF|ACCESS_CONF,
                    "SetEnvIfPlus"
                    " <attribute> <regex> [!]<env-variable>[=<value>] [late],"
                    " defines an expression to apply to the specified request"
                    " attribute (request header) and sets the defined"
                    " environment variable if the expression matches."),
  AP_INIT_TAKE_ARGV("SetEnvIfPlusNoCase", sp_std_cmd, SP_ICASE_MAGIC,
                    RSRC_CONF|ACCESS_CONF,
                    "SetEnvIfPlusNoCase"
                    " <attribute> <regex> [!]<env-variable>[=<value>] [late]"
                    " defines an expression to apply to the specified request"
                    " attribute (request header) and sets the defined"
                    " environment variable if the expression matches."),
  AP_INIT_TAKE3("SetQueryIfPlus", sp_std_req_query_cmd, NULL,
                RSRC_CONF|ACCESS_CONF,
                "SetQueryIfPlus <attribute> <regex> <query>[=<value>]"
                " defines an expression to apply to the specified request"
                " attribute (request header) and sets the defined"
                " request query paramater if the expression matches."),
  AP_INIT_TAKE1("ChangeQueryIfPlus", sp_std_req_changequery_cmd, NULL,
                RSRC_CONF|ACCESS_CONF,
                "CahngeQueryIfPlus <env-variable>"
                " replaces the request query by the value of the"
                " specifed environment variable."),
  AP_INIT_TAKE_ARGV("ResponseSetEnvIfPlus", sp_res_std_cmd, NULL,
                    ACCESS_CONF,
                    "ResponseSetEnvIfPlus"
                    " <attribute> <regex> [!]<env-variable>[=<value>],"
                    " defines an expression to apply to the specified response"
                    " attribute (response header) and sets the defined"
                    " environment variable if the expression matches."),
  AP_INIT_TAKE_ARGV("ResponseSetEnvIfPlusNoCase", sp_res_std_cmd, SP_ICASE_MAGIC,
                    ACCESS_CONF,
                    "ResponseSetEnvIfPlusNoCase"
                    " <attribute> <regex> [!]<env-variable>[=<value>]"
                    " defines an expression to apply to the specified response"
                    " attribute (response header) and sets the defined"
                    " environment variable if the expression matches."),
  AP_INIT_TAKE_ARGV("RequestHeaderPlus", sp_reqheader_cmd, NULL,
                    RSRC_CONF|ACCESS_CONF,
                    "RequestHeaderPlus set|unset|add <header> [<value>] [env=[!]<env-variable>] [late],"
                    " manipulates HTTP request headers. The first argument" 
                    " defines the action to take while the second argument"
                    " specifies the header name."
                    " The 'env=' attributes"
                    " is used to define additional conditions (variable must be set)."
                    " The 'late' attribute enforces late processing similar to"
                    " the SetEnvIfPlus directive."),
  AP_INIT_TAKE_ARGV("HeaderPlus", sp_resheader_cmd, NULL,
                    RSRC_CONF|ACCESS_CONF,
                    "HeaderPlus directive is deprecated (use RequestHeaderPlus)."),
  AP_INIT_TAKE_ARGV("ResponseHeaderPlus", sp_resheader_cmd, NULL,
                    RSRC_CONF|ACCESS_CONF,
                    "ResponseHeaderPlus set|unset|add <header> [<value>] [env=[!]<env-variable>],"
                    " manipulates HTTP request headers. The first argument" 
                    " defines the action to take while the second argument"
                    " specifies the header name."
                    " The 'env=' attributes"
                    " is used to define additional conditions (variable must be set)."),
  AP_INIT_TAKE3("ChangeRequestHeaderPlus", sp_changereqheader_cmd, NULL,
                RSRC_CONF|ACCESS_CONF,
                "ChangeRequestHeaderPlus <header> <regex> <value>, changes the"
                " value of the specified request header if the regex matches."),
  AP_INIT_TAKE2("RequestHeaderRemovePattern", sp_removepattern_cmd, NULL,
                ACCESS_CONF,
                "RequestHeaderRemovePattern <header> <pattern>, removes the"
                " specified pattern from the HTTP request header."),
  AP_INIT_TAKE3("ChangeResponseHeaderPlus", sp_changeresheader_cmd, NULL,
                RSRC_CONF|ACCESS_CONF,
                "ChangeResponseHeaderPlus <header> <regex> <value>, changes the"
                " value of the specified response header if the regex matches."),
  AP_INIT_TAKE_ARGV("SetEnvIfCmpPlus", sp_cmp_cmd, NULL,
                    RSRC_CONF|ACCESS_CONF,
                    "SetEnvIfCmpPlus <env-variable1> eq|ne|gt|lt <env-variable2> [!]<env-variable>[=<value>] [late],"
                    " sets the specifed environment variable if the specifed env-variables"
                    " are alphabetically or numerical equal (eq), not equal (ne),"
                    " greater (gt), less (lt). The value string may also contain other"
                    " environment variable names surrounded by \""SP_VAR_PFX"\" and \"}\" which"
                    " are resolved by the values of those variables."),
  AP_INIT_TAKE3("SetUserPlus", sp_setuser_cmd, NULL,
                ACCESS_CONF,
                "SetUserPlus <header> <regex> <value>,"
                " used to set the an authenticated user name using the 'value'."),
  AP_INIT_TAKE2("SetStatusPlus", sp_status_cmd, NULL,
                RSRC_CONF|ACCESS_CONF,
                "SetStatusPlus <code> <env-variable>,"
                " changes the status code of the HTTP response if the specified"
                " environment variable is set."),
  // TODO support per-directory config
  AP_INIT_TAKE1("CookieEncPlus", sp_cookie_cmd, NULL,
                RSRC_CONF,
                "CookieEncPlus <name>,"
                " encrypts the value of the specified cookies. The passphrase for"
                " encryption is defined by the "SP_COOKIE_KEY" variable."),
  AP_INIT_TAKE2("SetHashHeaderPlus", sp_hash_cmd, NULL,
                RSRC_CONF,
                "SetHashHeaderPlus <header name> <env-variable>, creates a hash"
                " of the value stored in the defined variable and stores it"
                " within the defined requst header."),
  AP_INIT_TAKE2("AddOutputFilterPlus", sp_outfilter_cmd, NULL,
                RSRC_CONF|ACCESS_CONF,
                "AddOutputFilterPlus <filter name> [!]<env-variable>,"
                " adds the ouput filter to the response if"
                " the specified environment variable is set."),
  AP_INIT_TAKE1("SetVarPrefixPlus", sp_var_pfx_cmd, NULL,
                RSRC_CONF,
                "SetVarPrefixPlus '$'|'%', defines the prefix of the"
                " variable place holder which is either ${...} or"
                " %{...} for compatibilty with Apache 2.4."
                " Default is '$'."),
  { NULL }
};

/************************************************************************
 * apache register 
 ***********************************************************************/
static void sp_register_hooks(apr_pool_t * p) {
  static const char *pre[] = { "mod_setenvif.c", "mod_ssl.c", "mod_session.c", NULL };
  static const char *post[] = { "mod_headers.c", NULL };
  ap_hook_post_read_request(sp_post_read_request, pre, post, APR_HOOK_MIDDLE);
  ap_hook_header_parser(sp_header_parser, pre, post, APR_HOOK_MIDDLE);
  ap_hook_check_user_id(sp_user_id, pre, NULL, APR_HOOK_FIRST);
  ap_hook_fixups(sp_fixup, pre, post, APR_HOOK_MIDDLE);
  ap_hook_post_config(sp_post_config, pre, NULL, APR_HOOK_MIDDLE);
#ifdef SP_INTERNAL_TEST
  ap_hook_handler(sp_handler, NULL, NULL, APR_HOOK_MIDDLE);
#endif
  ap_register_output_filter("sp_out_filter", sp_out_filter, NULL, AP_FTYPE_RESOURCE);
  ap_register_output_filter("sp_out_err_filter", sp_out_err_filter, NULL, AP_FTYPE_RESOURCE);
  ap_hook_insert_filter(sp_insert_filter, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_insert_error_filter(sp_insert_err_filter, NULL, NULL, APR_HOOK_MIDDLE);
}

/************************************************************************
 * apache module definition 
 ***********************************************************************/
module AP_MODULE_DECLARE_DATA setenvifplus_module ={ 
  STANDARD20_MODULE_STUFF,
  sp_dir_config_create,                    /**< dir config creater */
  sp_dir_config_merge,                     /**< dir merger */
  sp_srv_config_create,                    /**< server config */
  sp_srv_config_merge,                     /**< server merger */
  sp_config_cmds,                          /**< command table */
  sp_register_hooks,                       /**< hook registery */
};
